from enum import Flag, auto

from .norm_data import PROCLITICS, ENCLITICS, ELISION, MOVABLE
from .utils import (
    strip_last_accent, grave_to_acute, strip_last_accent_if_two, count_accents
)

ENCLITICS_NORM = {
    strip_last_accent(word): word
    for word in ENCLITICS
}


class Norm(Flag):
    UNCHANGED = 0
    GRAVE = auto()
    ELISION = auto()
    MOVABLE = auto()
    EXTRA = auto()
    CAPITALISED = auto()
    PROCLITIC = auto()
    ENCLITIC = auto()
    NO_ACCENT = auto()
    ALL = (
        GRAVE | ELISION | MOVABLE | EXTRA | CAPITALISED |
        PROCLITIC | ENCLITIC | NO_ACCENT
    )


class Normaliser:

    def __init__(self, config=Norm.ALL, proper_nouns=set()):
        self.config = config
        self.proper_nouns = proper_nouns

    def normalise(self, token):

        norm = token
        norm_code = Norm.UNCHANGED

        if self.config & Norm.GRAVE:
            # change graves to acutes
            temp = grave_to_acute(norm)
            if norm != temp:
                norm_code |= Norm.GRAVE
            norm = temp

        if self.config & Norm.ELISION:
            if norm in ELISION:
                norm = ELISION[norm]
                norm_code |= Norm.ELISION

        if self.config & Norm.MOVABLE:
            if norm in MOVABLE:
                norm = MOVABLE[norm]
                norm_code |= Norm.MOVABLE

        if self.config & Norm.EXTRA:
            # strip last accent if two
            temp = strip_last_accent_if_two(norm)
            if norm != temp:
                norm_code |= Norm.EXTRA
            norm = temp

        if self.config & Norm.CAPITALISED:
            if norm not in self.proper_nouns:
                if norm != norm.lower():
                    norm = norm.lower()
                    norm_code |= Norm.CAPITALISED

        if count_accents(norm) == 0:
            if norm.lower() in PROCLITICS:
                if self.config & Norm.PROCLITIC:
                    norm = norm.lower()
                    norm_code |= Norm.PROCLITIC
            elif norm.lower() in ENCLITICS_NORM:
                if self.config & Norm.ENCLITIC:
                    norm = ENCLITICS_NORM[norm.lower()]
                    norm_code |= Norm.ENCLITIC
            else:
                if self.config & Norm.NO_ACCENT:
                    norm_code |= Norm.NO_ACCENT

        return norm, norm_code



#!/usr/bin/env python3

import fileinput
import sys

from .utils import nfc, nfd, convert_to_2019


def convert(func):
    lines_changed = 0

    with fileinput.input() as f:
        for line in f:
            text = func(line)
            print(text, end="")
            if text != line:
                lines_changed += 1

    print(f"{lines_changed} lines changed", file=sys.stderr)


def to_nfc():
    convert(nfc)


def to_nfd():
    convert(nfd)


def to_2019():
    convert(convert_to_2019)



PROCLITICS = {
    "ὁ", "ἡ", "οἱ", "αἱ",
    "ἐν", "εἰς", "ἐξ", "ἐκ",
    "εἰ", "ὡς",
    "οὐ", "οὐκ", "οὐχ",
}


ENCLITICS = {
    # personal pronouns
    "μου", "μοι", "με",
    "σου", "σοι", "σε",

    # indefinite pronouns
    "τὶς", "τὶ", "τινός", "τινί", "τινά", "τινές", "τινάς", "τινῶν", "τισίν",
    "τισί",

    # indefinite adverbs
    "πού", "ποτέ", "πώ", "πώς",

    # dissyllabic forms of εἰμί
    "εἰμί", "εἰσίν", "εἰσί", "ἐσμέν", "ἐστέ", "ἐστίν", "ἐστί",

    # dissyllabic forms of φημί
    "φησίν", "φημί", "φασίν",

    # certain particles
    "γέ", "τέ", "τοι",
}


ELISION = {
    "ἀλλ’": "ἀλλά",
    "ἀνθ’": "ἀντί",
    "ἀπ’": "ἀπό",
    "ἀφ’": "ἀπό",
    "γ’": "γε",
    "γένοιτ’": "γένοιτο",
    "δ’": "δέ",
    "δεῦρ’": "δεῦρο",
    "δι’": "διά",
    "δύναιτ’": "δύναιτο",
    "εἶτ’": "εἶτα",
    "ἐπ’": "ἐπί",
    "ἔτ’": "ἔτι",
    "ἐφ’": "ἐπί",
    "ἡγοῖντ’": "ἡγοῖντο",
    "ἵν’": "ἵνα",
    "καθ’": "κατά",
    "κατ’": "κατά",
    "μ’": "με",
    "μεθ’": "μετά",
    "μετ’": "μετά",
    "μηδ’": "μηδέ",
    "μήδ’": "μηδέ",  # @@@
    "ὅτ’": "ὅτε",
    "οὐδ’": "οὐδέ",
    "πάνθ’": "πάντα",
    "πάντ’": "πάντα",
    "παρ’": "παρά",
    "ποτ’": "ποτε",
    "σ’": "σε",
    "τ’": "τε",
    "ταῦθ’": "ταῦτα",
    "ταῦτ’": "ταῦτα",
    "τοῦτ’": "τοῦτο",
    "ὑπ’": "ὑπό",
    "ὑφ’": "ὑπό",
    # "ἔσθ’": "???",
    # "θεῖ’": "???",
}

MOVABLE = {
    "ἐξ": "ἐκ",

    "οὐκ": "οὐ",
    "οὐχ": "οὐ",
}




ELLIPSIS = r"\u2026"
LATIN = r"[A-Za-zë]+[\u002C\u002E\u003A\u003B\u003F]?"
NUMBER = r"[αβγδεϛζηθικλμνξοπϟρστυφχψωϡ]+[\u02B9]"
PUNC = r"[\u002C\u002E\u003B\u00B7]"
GRC_CHAR = r"[\u0374\u0390-\u03A1\u03A3-\u03C1\u03C3-\u03CE" \
         r"\u1F00-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57" \
         r"\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC" \
         r"\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC" \
         r"\u1FF2-\u1FF4\u1FF6-\u1FFC]"
GRC_WORD = fr"{GRC_CHAR}+\u03C2?\u2019?"

NFD_TOKEN = r"""
[
    \u0028              # (
    \u00AB              # «
]?
(?P<word>
    [
        \u0300          # combining grave
        \u0301          # combining acute
        \u0304          # combining macron
        \u0308          # combining diaeresis
        \u0313          # combining smooth breathing
        \u0314          # combining rough breathing
        \u0342          # combining circumflex
        \u0345          # combining iota subscript
        \u0391-\u03A9   # uppercase
        \u03B1-\u03C9   # lowercase
        \u002D          # - (hyphen)
    ]+
[
    \u2019              # ’ (apostrophe)
]?
)
[
    \u0021              # !
    \u0029              # )
    \u002C              # ,
    \u002E              # .
    \u003B              # ;
    \u00B7              # · (middle dot)
    \u00BB              # »
]*
$
"""



import unicodedata


VARIA = "\u0300"
OXIA = "\u0301"
PERISPOMENI = "\u0342"

PSILI = "\u0313"
DASIA = "\u0314"

DIAERESIS = "\u0308"

BREATHING = [PSILI, DASIA]
ACCENTS = [VARIA, OXIA, PERISPOMENI]


def nfd(s):
    return unicodedata.normalize("NFD", s)


def nfc(s):
    return unicodedata.normalize("NFC", s)


def nfkc(s):
    return unicodedata.normalize("NFKC", s)


def convert_to_2019(s):
    return s.replace("\u02BC", "\u2019").replace("\u1FBF", "\u2019")


def strip_accents(s):
    return nfc("".join(
        cp for cp in nfd(s) if cp not in ACCENTS
    ))


def count_accents(word):
    count = 0
    for c in nfd(word):
        if c in ACCENTS:
            count += 1
    return count


def strip_last_accent(word):
    x = list(word)
    for i, ch in enumerate(x[::-1]):
        s = strip_accents(ch)
        if s != ch:
            x[-i - 1] = s
            break
    return "".join(x)


def grave_to_acute(word):
    return nfc("".join(
        (OXIA if cp == VARIA else cp) for cp in nfd(word)
    ))


def strip_last_accent_if_two(word):
    if count_accents(word) == 2:
        norm = strip_last_accent(word)
    else:
        norm = word
    return norm


def breathing_check(word):
    # note: doesn't check for mid-word breathing marks
    d = [ch for ch in nfd(word.lower()) if ch not in ACCENTS]
    if d[0] in "αεηιοω":
        if len(d) > 1:
            if d[1] in BREATHING:
                if len(d) > 2 and d[2] in "ιυ":
                    if len(d) > 3 and d[3] == DIAERESIS:
                        return True
                    else:
                        return False
                else:
                    return True
            elif d[1] in "ιυ":
                if len(d) > 2 and d[2] in BREATHING:
                    pass
                else:
                    return False
            else:
                return False
        else:
            return False
    return True



from setuptools import setup


with open("README.md", "r") as f:
    long_description = f.read()


setup(
    name="greek-normalisation",
    version="0.5.1",
    description="Python 3 utilities for validating and normalising Ancient Greek text",
    url="http://github.com/jtauber/greek-normalisation",
    author="James Tauber",
    author_email="jtauber@jtauber.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=["greek_normalisation"],
    entry_points={
        "console_scripts": [
            "to2019 = greek_normalisation.convert_files:to_2019",
            "toNFC = greek_normalisation.convert_files:to_nfc",
            "toNFD = greek_normalisation.convert_files:to_nfd",
        ],
    },
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Text Processing :: Linguistic",
    ],
)
